from collections import defaultdict, namedtuple
import torch

# When using the sliding window trick for long sequences,
# we take the representation of each token with maximal context.
# Take average of the BERT embeddings of these BPE sub-tokens
# as the embedding for the word.
# Take *weighted* average of the word embeddings through all layers.

def extract_bert_hidden_states(all_encoder_layers, max_doc_len, features, weighted_avg=False):
  # assert all_encoder_layers.requires_grad == False
  num_layers, batch_size, num_chunk, max_token_len, bert_dim = all_encoder_layers.shape
  out_features = torch.Tensor(num_layers, batch_size, max_doc_len, bert_dim).fill_(0)
  device = all_encoder_layers.get_device() if all_encoder_layers.is_cuda else None
  if device is not None:
    out_features = out_features.to(device)

  token_count = []
  # Map BERT tokens to doc words
  for i, ex_feature in enumerate(features): # Example
    ex_token_count = defaultdict(int)
    for j, chunk_feature in enumerate(ex_feature): # Chunk
      for k in chunk_feature.token_is_max_context: # Token
        if chunk_feature.token_is_max_context[k]:
          doc_word_idx = chunk_feature.token_to_orig_map[k]
          out_features[:, i, doc_word_idx] += all_encoder_layers[:, i, j, k]
          ex_token_count[doc_word_idx] += 1
    token_count.append(ex_token_count)

  for i, ex_token_count in enumerate(token_count):
    for doc_word_idx, count in ex_token_count.items():
      out_features[:, i, doc_word_idx] /= count

  # Average through all layers
  if not weighted_avg:
    out_features = torch.mean(out_features, 0)
  return out_features

def convert_text_to_bert_features(text, bert_tokenizer, max_seq_length, doc_stride):
  # The convention in BERT is:
      # (a) For sequence pairs:
      #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
      #  type_ids: 0   0  0    0    0     0       0 0    1  1  1  1   1 1
      # (b) For single sequences:
      #  tokens:   [CLS] the dog is hairy . [SEP]
      #  type_ids: 0   0   0   0  0     0 0

  tok_to_orig_index = []
  all_doc_tokens = []
  for (i, token) in enumerate(text):
    sub_tokens = bert_tokenizer.wordpiece_tokenizer.tokenize(token.lower())
    for sub_ in sub_tokens:
      tok_to_orig_index.append(i)
      all_doc_tokens.append(sub_)

  # The -2 accounts for [CLS] and [SEP]
  max_tokens_for_doc = max_seq_length - 2

  # We can have documents that are longer than the maximum sequence length.
  # To deal with this we do a sliding window approach, where we take chunks
  # of the up to our max length with a stride of `doc_stride`.
  _DocSpan = namedtuple(  # pylint: disable=invalid-name
      "DocSpan", ["start", "length"])
  doc_spans = []
  start_offset = 0
  while start_offset < len(all_doc_tokens):
    length = len(all_doc_tokens) - start_offset
    if length > max_tokens_for_doc:
      length = max_tokens_for_doc
    doc_spans.append(_DocSpan(start=start_offset, length=length))
    if start_offset + length == len(all_doc_tokens):
      break
    start_offset += min(length, doc_stride)

  out_features = []
  for (doc_span_index, doc_span) in enumerate(doc_spans):
    tokens = []
    token_to_orig_map = {}
    token_is_max_context = {}
    segment_ids = []
    tokens.append("[CLS]")
    segment_ids.append(0)

    for i in range(doc_span.length):
      split_token_index = doc_span.start + i
      token_to_orig_map[len(tokens)] = tok_to_orig_index[split_token_index]

      is_max_context = _check_is_max_context(doc_spans, doc_span_index,
                                             split_token_index)
      token_is_max_context[len(tokens)] = is_max_context
      tokens.append(all_doc_tokens[split_token_index])
      segment_ids.append(0)
    tokens.append("[SEP]")
    segment_ids.append(0)

    input_ids = bert_tokenizer.convert_tokens_to_ids(tokens)

    # The mask has 1 for real tokens and 0 for padding tokens. Only real
    # tokens are attended to.
    input_mask = [1] * len(input_ids)

    feature = BertInputFeatures(
    doc_span_index=doc_span_index,
    tokens=tokens,
    token_to_orig_map=token_to_orig_map,
    token_is_max_context=token_is_max_context,
    input_ids=input_ids,
    input_mask=input_mask,
    segment_ids=segment_ids)
    out_features.append(feature)
  return out_features


def _check_is_max_context(doc_spans, cur_span_index, position):
  """Check if this is the 'max context' doc span for the token."""

  # Because of the sliding window approach taken to scoring documents, a single
  # token can appear in multiple documents. E.g.
  #  Doc: the man went to the store and bought a gallon of milk
  #  Span A: the man went to the
  #  Span B: to the store and bought
  #  Span C: and bought a gallon of
  #  ...
  #
  # Now the word 'bought' will have two scores from spans B and C. We only
  # want to consider the score with "maximum context", which we define as
  # the *minimum* of its left and right context (the *sum* of left and
  # right context will always be the same, of course).
  #
  # In the example the maximum context for 'bought' would be span C since
  # it has 1 left context and 3 right context, while span B has 4 left context
  # and 0 right context.
  best_score = None
  best_span_index = None
  for (span_index, doc_span) in enumerate(doc_spans):
    end = doc_span.start + doc_span.length - 1
    if position < doc_span.start:
      continue
    if position > end:
      continue
    num_left_context = position - doc_span.start
    num_right_context = end - position
    score = min(num_left_context, num_right_context) + 0.01 * doc_span.length
    if best_score is None or score > best_score:
      best_score = score
      best_span_index = span_index

  return cur_span_index == best_span_index

class BertInputFeatures(object):
    """A single set of BERT features of data."""
    def __init__(self,
                 doc_span_index,
                 tokens,
                 token_to_orig_map,
                 token_is_max_context,
                 input_ids,
                 input_mask,
                 segment_ids):
        self.doc_span_index = doc_span_index
        self.tokens = tokens
        self.token_to_orig_map = token_to_orig_map
        self.token_is_max_context = token_is_max_context
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids



"""
Module to handle universal/general constants used across files.
"""

################################################################################
# Constants #
################################################################################

# GENERAL CONSTANTS:
VERY_SMALL_NUMBER = 1e-31
INF = 1e20


_PAD_TOKEN = '#pad#'
_UNK_TOKEN = '<unk>'
_SOS_TOKEN = '<s>'
_EOS_TOKEN = '</s>'


# LOG FILES ##
_CONFIG_FILE = 'config.json'
_SAVED_WEIGHTS_FILE = 'params.saved'
_SAVED_METRICS_FILE = 'metric.log'
# _PREDICTION_FILE = "test_pred.txt"
# _REFERENCE_FILE = "test_ref.txt"



import collections
from torch import nn
import torch as th
from ..base import NodeClassifierBase
from .FeedForwardNNLayer import FeedForwardNNLayer

class FeedForwardNN(NodeClassifierBase):
    r"""Specific class for node classification task.


    Parameters
    ----------

    input_size : int 
                 The length of input node embeddings
    num_class : int 
               The number of node catrgoriey for classification
    hidden_size : list of int type values
                  Example for two layers's FeedforwardNN: [50, 20]
    activation: the activation function class for each fully connected layer
                Default: nn.ReLU()
                Example: nn.ReLU(),nn.Sigmoid().        

    """     
    def __init__(self, input_size, num_class, hidden_size, activation=nn.ReLU()):        
        super(FeedForwardNN, self).__init__()
        
        
        self.classifier=FeedForwardNNLayer(input_size, num_class, hidden_size, activation)

    def forward(self, input_graph):
        r"""
        Forward functions to compute the logits tensor for node classification.
    
      
    
        Parameters
        ----------
    
        input graph : GraphData
                     The tensors stored in the node feature field named "node_emb"  in the 
                     input_graph are used  for classification.

    
        Returns 
        ---------
        
        output_graph : GraphData
                      The computed logit tensor for each nodes in the graph are stored
                      in the node feature field named "node_logits".
                      logit tensor shape is: [num_class] 
        """ 

        node_emb=input_graph.node_features['node_emb']
        input_graph.node_features['logits']=self.classifier(node_emb)
        
        return input_graph






from setuptools import setup, find_packages


def get_cuda_version():
    import os
    nvsmi = os.popen('nvidia-smi')
    gpu_info_str = nvsmi.read()
    if gpu_info_str == '':
        return ''
    else:
        cuda_version_prompt = 'CUDA Version: '
        version_number_start = gpu_info_str.find(cuda_version_prompt) + len(cuda_version_prompt)
        version_number = gpu_info_str[version_number_start:].split(' ')[0]
        return '-cu{}'.format(version_number.replace('.', ''))
        # install_requirement.append('dgl-cuda{} >= 0.4'.format(version_number))


if __name__ == '__main__':
    install_requirement = ['torch >= 1.6.0', 'pythonds', 'nltk >= 3.5', 'stanfordcorenlp', 'scipy >= 1.5.2',
                           'scikit-learn >= 0.23.2', 'networkx >= 2.5', 'dgl{} >= 0.4'.format(get_cuda_version())]
    setup(
        name='graph4nlp{}'.format(get_cuda_version()),
        version='0.1a100011',
        description='GNN for NLP library',
        author='graph4nlp team',
        license='MIT',
        packages=find_packages('.', exclude=("examples.*", "examples", "graph4nlp.pytorch.test.*", "graph4nlp.pytorch.test")),
        install_requires=install_requirement,
    )



import json, re, string
from unicodedata import normalize
import argparse
import tqdm


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--train_raw_dataset_src', default='/home/shiina/shiina/lib/dataset/europarl-v7.fr-en.fr', type=str, help='path to the config file')
    parser.add_argument('--train_raw_dataset_tgt', type=str, default='/home/shiina/shiina/lib/dataset/europarl-v7.fr-en.en', help='rank')
    parser.add_argument('--train_output_dataset', type=str, default='/home/shiina/shiina/lib/dataset/raw/train.json', help="raw data path")

    parser.add_argument('--val_raw_dataset_src', default='/home/shiina/shiina/lib/dataset/test2011/newstest2011.fr',
                        type=str, help='path to the config file')
    parser.add_argument('--val_raw_dataset_tgt', type=str,
                        default='/home/shiina/shiina/lib/dataset/test2011/newstest2011.en', help='rank')
    parser.add_argument('--val_output_dataset', type=str, default='/home/shiina/shiina/lib/dataset/raw/test.json',
                        help="raw data path")

    cfg = parser.parse_args()
    return cfg



class TextProcessor:
    def __init__(self, raw_dataset_src, raw_dataset_tgt, output_dataset):
        self.raw_dataset_src = raw_dataset_src
        self.raw_dataset_tgt = raw_dataset_tgt
        self.output_dataset = output_dataset

    def _preprocess(self, lines):

        # clean a list of lines
        cleaned = list()
        # prepare regex for char filtering
        re_print = re.compile('[^%s]' % re.escape(string.printable))
        # prepare translation table for removing punctuation
        table = str.maketrans('', '', string.punctuation)
        for line in lines:
            # normalize unicode characters
            line = normalize('NFD', line).encode('ascii', 'ignore')
            line = line.decode('UTF-8')
            # tokenize on white space
            line = line.split()
            # convert to lower case
            line = [word.lower() for word in line]
            # remove punctuation from each token
            line = [word.translate(table) for word in line]
            # remove non-printable chars form each token
            line = [re_print.sub('', w) for w in line]
            # remove tokens with numbers in them
            line = [word for word in line if word.isalpha()]
            # store as string
            cleaned.append(' '.join(line))
        return cleaned

    def run(self):
        src_lines = self._read(self.raw_dataset_src)
        tgt_lines = self._read(self.raw_dataset_tgt)
        print(len(src_lines), len(tgt_lines))
        assert len(src_lines) == len(tgt_lines)
        output_content = []
        for src, tgt in zip(src_lines, tgt_lines):
            if src.strip() == "":
                continue
            processed = self._preprocess([src, tgt])
            output_content.append((processed[0], processed[1]))
        with open(self.output_dataset, "w") as f:
            json.dump(output_content, f)
        print("done")

    def _read(self, path):
        with open(path, "r") as f:
            content = f.read()
        sentences = content.strip().split("\n")
        print(sentences[0])
        return sentences

if __name__ == "__main__":
    opt = get_args()
    # process training set
    # processor = TextProcessor(raw_dataset_src=opt.train_raw_dataset_src, raw_dataset_tgt=opt.train_raw_dataset_tgt,
    #                           output_dataset=opt.train_output_dataset)
    # processor.run()
    # process validation set
    processor = TextProcessor(raw_dataset_src=opt.val_raw_dataset_src, raw_dataset_tgt=opt.val_raw_dataset_tgt,
                              output_dataset=opt.val_output_dataset)
    processor.run()







from torch import nn
import torch
from ..base import KGCompletionLayerBase
import torch.nn.functional as F


class TransELayer(KGCompletionLayerBase):
    r"""Specific class for knowledge graph completion task.
    TransE from paper `Translating Embeddings for Modeling
    Multi-relational Data <https://papers.nips.cc/paper/5071
    -translating-embeddings-for-modeling-multi-relational-data.pdf>`__.

    .. math::
        f(s, r, o) & = ||e_s + w_r - e_o||_p

    Parameters
    ----------
    p_norm: int
        Default: 1

    rel_emb_from_gnn: bool
        If `rel_emb` is computed from GNN, rel_emb_from_gnn is set to `True`.
        Else, rel_emb is initialized as nn.Embedding randomly. Default: `True`.

    num_relations: int
        Number of relations. `num_relations` is needed if rel_emb_from_gnn==True.
        Default: `None`.

    embedding_dim: int
        Dimension of the rel_emb. `embedding_dim` is needed if rel_emb_from_gnn==True.
        Default: `0`.

    loss_name: str
        The loss type selected fot the KG completion task.

    """

    def __init__(self,
                 p_norm=1,
                 rel_emb_from_gnn=True,
                 num_relations=None,
                 embedding_dim=None,
                 loss_name='BCELoss'):
        super(TransELayer, self).__init__()
        self.p_norm = p_norm
        self.rel_emb_from_gnn = rel_emb_from_gnn
        if self.rel_emb_from_gnn == False:
            assert num_relations != None
            assert embedding_dim != None
            self.rel_emb = nn.Embedding(num_relations, embedding_dim)
            self.reset_parameters()
        self.loss_name = loss_name
        self.reset_parameters()

    def reset_parameters(self):
        if self.rel_emb_from_gnn == False:
            nn.init.xavier_normal_(self.rel_emb.weight.data)

    def forward(self,
                node_emb,
                rel_emb=None,
                list_e_r_pair_idx=None,
                list_e_e_pair_idx=None,
                multi_label=None):
        r"""

        Parameters
        ----------

        node_emb: tensor [N,H]
            N: number of nodes in the whole KG graph
            H: length of the node embeddings (entity embeddings)

        rel_emb: tensor [N_r,H]
            N: number of relations in the whole KG graph
            H: length of the relation embeddings

        list_e_r_pair_idx: list of tuple
            a list of index of head entities and relations that needs
            predicting the tail entities between them. Default: `None`

        list_e_e_pair_idx: list of tuple
            a list of index of head entities and tail entities that needs
            predicting the relations between them. Default: `None`.
            Only one of `list_e_r_pair_idx` and `list_e_e_pair_idx` can be `None`.

        multi_label: tensor [L, N]
            multi_label is a binary matrix. Each element can be equal to 1 for true label
            and 0 for false label (or 1 for true label, -1 for false label).
            multi_label[i] represents a multi-label of a given head-rel pair or head-tail pair.
            L is the length of list_e_r_pair_idx, list_e_e_pair_idx or batch size.
            N: number of nodes in the whole KG graph.


        Returns
        -------
        pred: tensor [L, N]
            logit tensor. [L, N] The score logits for all nodes preidcted.

        pred_pos: tensor [L_p]
            The predition scores of positive examples.

        pred_neg: tensor [L_n]
            The predition scores of negative examples. L_p + L_n == L * N.

        """
        if self.rel_emb_from_gnn == False:
            assert rel_emb is None
            rel_emb = self.rel_emb.weight

        if list_e_r_pair_idx is None and list_e_e_pair_idx is None:
            raise RuntimeError("Only one of `list_e_r_pair_idx` and `list_e_e_pair_idx` can be `None`.")

        assert node_emb.size()[1] == rel_emb.size()[1]

        if list_e_r_pair_idx is not None:
            ent_idxs = torch.LongTensor([x[0] for x in list_e_r_pair_idx])
            rel_idxs = torch.LongTensor([x[1] for x in list_e_r_pair_idx])

            selected_ent_embs = node_emb[ent_idxs].squeeze()  # [L, H]. L is the length of list_e_r_pair_idx
            selected_rel_embs = rel_emb[rel_idxs].squeeze()  # [L, H]. L is the length of list_e_r_pair_idx

            selected_ent_embs = F.normalize(selected_ent_embs, 2, -1)
            selected_rel_embs = F.normalize(selected_rel_embs, 2, -1)
            node_emb = F.normalize(node_emb, 2, -1)

            head_add_rel = selected_ent_embs + selected_rel_embs  # [L, H]
            head_add_rel = head_add_rel.view(head_add_rel.size()[0], 1, head_add_rel.size()[1])  # [L, 1, H]
            head_add_rel = head_add_rel.repeat(1, node_emb.size()[0], 1)

            node_emb = node_emb.view(1, node_emb.size()[0], node_emb.size()[1])  # [1, N, H]
            node_emb = node_emb.repeat(head_add_rel.size()[0], 1, 1)

            result = head_add_rel - node_emb  # head+rel-tail [L, N, H]

        elif list_e_e_pair_idx is not None:
            ent_head_idxs = torch.LongTensor([x[0] for x in list_e_e_pair_idx])
            ent_tail_idxs = torch.LongTensor([x[1] for x in list_e_e_pair_idx])

            selected_ent_head_embs = node_emb[ent_head_idxs].squeeze()  # [L, H]. L is the length of list_e_e_pair_idx
            selected_ent_tail_embs = rel_emb[ent_tail_idxs].squeeze()  # [L, H]. L is the length of list_e_e_pair_idx

            selected_ent_head_embs = F.normalize(selected_ent_head_embs, 2, -1)
            selected_ent_tail_embs = F.normalize(selected_ent_tail_embs, 2, -1)
            rel_emb = F.normalize(rel_emb, 2, -1)

            head_sub_tail = selected_ent_head_embs - selected_ent_tail_embs  # [L, H]
            head_sub_tail = head_sub_tail.view(head_sub_tail.size()[0], 1, head_sub_tail.size()[1])  # [L, 1, H]
            head_sub_tail = head_sub_tail.repeat(1, rel_emb.size()[0], 1)  # [L, N, H]

            rel_emb = rel_emb.view(1, rel_emb.size()[0], rel_emb.size()[1])  # [1, N, H]
            rel_emb = rel_emb.repeat(head_sub_tail.size()[0], 1, 1)  # [L, N, H]

            result = head_sub_tail + rel_emb  # head-tail+rel [L, N, H]

        if self.loss_name in ['SoftMarginLoss']:
            # target labels are numbers selecting from -1 and 1.
            pred = torch.norm(result, self.p_norm, dim=2)  # TODO
        else:
            pred = torch.softmax(torch.norm(result, self.p_norm, dim=2), dim=-1)  # logits [L, N]

        if multi_label is not None:
            idxs_pos = torch.nonzero(multi_label == 1.)
            pred_pos = pred[idxs_pos[:, 0], idxs_pos[:, 1]]

            idxs_neg = torch.nonzero(multi_label == 0.)
            pred_neg = pred[idxs_neg[:, 0], idxs_neg[:, 1]]
            return pred, pred_pos, pred_neg
        else:
            return pred


from graph4nlp.pytorch.data.dataset import Text2TextDataItem, Text2TextDataset
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
import torch
import os
import json
from stanfordcorenlp import StanfordCoreNLP


class EuroparlNMTDataset(Text2TextDataset):
    def __init__(self, root_dir, topology_builder, topology_subdir=None, graph_type='static',
                 edge_strategy=None, merge_strategy='tailhead', **kwargs):
        super(EuroparlNMTDataset, self).__init__(root_dir=root_dir, topology_builder=topology_builder, share_vocab=False,
                                          topology_subdir=topology_subdir, graph_type=graph_type,
                                          edge_strategy=edge_strategy, merge_strategy=merge_strategy, **kwargs)

    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'train.json', 'test': 'test.json'}

    @property
    def processed_file_names(self):
        """At least 3 reserved keys should be fiiled: 'vocab', 'data' and 'split_ids'."""
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        raise NotImplementedError("It shouldn't be called now")


    def parse_file(self, files):
        with open(files, "r") as f:
            datalist = json.load(f)
        data = []
        for item in datalist:
            if item[0].strip() == "" or item[1].strip() == "":
                continue
            dataitem = Text2TextDataItem(input_text=item[0], output_text=item[1], tokenizer=self.tokenizer, share_vocab=self.share_vocab)
            data.append(dataitem)
        if len(data) > 10000:
            data = data[0:10000]
        return data

    def build_topology(self, data_items):
        """
        Build graph topology for each item in the dataset. The generated graph is bound to the `graph` attribute of the
        DataItem.
        """
        if self.graph_type == 'static':
            print('Connecting to stanfordcorenlp server...')
            processor = StanfordCoreNLP('http://localhost', port=9000, timeout=1000)
            print('CoreNLP server connected.')
            processor_args = {
                'annotators': 'ssplit,tokenize,depparse',
                "tokenize.options":
                    "splitHyphenated=false,normalizeParentheses=false,normalizeOtherBrackets=false",
                "tokenize.whitespace": False,
                'ssplit.isOneSentence': False,
                'outputFormat': 'json'
            }
            cnt = 0
            for item in data_items:
                if cnt % 100 == 0:
                    print(cnt, len(data_items))
                cnt += 1
                graph = self.topology_builder.topology(raw_text_data=item.input_text,
                                                       nlp_processor=processor, processor_args=processor_args,
                                                       merge_strategy=self.merge_strategy,
                                                       edge_strategy=self.edge_strategy,
                                                       verbase=False)
                item.graph = graph
        elif self.graph_type == 'dynamic':
            for item in data_items:
                graph = self.topology_builder.raw_text_to_init_graph(item.input_text)
                item.graph = graph
        else:
            raise NotImplementedError('Currently only static and dynamic are supported!')


if __name__ == "__main__":
    dataset = EuroparlNMTDataset(root_dir="/home/shiina/shiina/lib/dataset",
                                 topology_builder=DependencyBasedGraphConstruction,
                                 topology_subdir='DependencyGraph')





import torch.nn as nn

from ..modules.loss.seq_generation_loss import SeqGenerationLoss


class Graph2SeqLoss(nn.Module):
    def __init__(self, ignore_index, use_coverage, coverage_weight):
        super(Graph2SeqLoss, self).__init__()
        self.loss_func = SeqGenerationLoss(ignore_index=ignore_index, use_coverage=use_coverage,
                                           coverage_weight=coverage_weight)

    def forward(self, logits, label, enc_attn_weights=None, coverage_vectors=None):
        loss = self.loss_func(logits=logits, label=label, enc_attn_weights=enc_attn_weights,
                              coverage_vectors=coverage_vectors)
        return loss


from operator import itemgetter
import pickle as pkl
import torch
from random import randint
import numpy as np
import copy


class Tree():
    def __init__(self):
        self.parent = None
        self.num_children = 0
        self.children = []

    def __str__(self, level=0):
        ret = ""
        for child in self.children:
            if isinstance(child, type(self)):
                ret += child.__str__(level+1)
            else:
                ret += "\t"*level + str(child) + "\n"
        return ret

    def add_child(self, c):
        if isinstance(c, type(self)):
            c.parent = self
        self.children.append(c)
        self.num_children = self.num_children + 1

    def to_string(self):
        r_list = []
        for i in range(self.num_children):
            if isinstance(self.children[i], Tree):
                r_list.append("( " + self.children[i].to_string() + " )")
            else:
                r_list.append(" "+str(self.children[i])+" ")
        return "".join(r_list)

    def to_text(self, form_manager):
        r_list = []
        for i in range(self.num_children):
            if isinstance(self.children[i], Tree):
                r_list.append("( " + self.children[i].to_text(form_manager) + " )")
            else:
                r_list.append(" "+form_manager.get_idx_symbol(self.children[i])+" ")
        return "".join(r_list)

    def to_list(self, tgt_vocab):
        r_list = []
        for i in range(self.num_children):
            if isinstance(self.children[i], type(self)):
                r_list.append(tgt_vocab.get_symbol_idx("("))
                cl = self.children[i].to_list(tgt_vocab)
                for k in range(len(cl)):
                    r_list.append(cl[k])
                r_list.append(tgt_vocab.get_symbol_idx(")"))
            else:
                r_list.append(self.children[i])
        return r_list

    @staticmethod
    def norm_tree(r_list, tgt_vocab):
        q = [Tree.convert_to_tree(r_list, 0, len(r_list), tgt_vocab)]
        head = 0
        while head < len(q):
            t = q[head]
            exchangeable_symbol_idx_list = tgt_vocab.get_symbol_idx_for_list(
                ['and', 'or'])
            if len(t.children) > 0 and (t.children[0] in exchangeable_symbol_idx_list):
                k = []
                for i in range(1, len(t.children)):
                    if isinstance(t.children[i], Tree):
                        k.append((t.children[i].to_string(), i))
                    else:
                        k.append((str(t.children[i]), i))
                sorted_t_dict = []
                k.sort(key=itemgetter(0))
                for key1 in k:
                    sorted_t_dict.append(t.children[key1[1]])

                for i in range(t.num_children-1):
                    t.children[i+1] = \
                        sorted_t_dict[i]
            for i in range(len(t.children)):
                if isinstance(t.children[i], Tree):
                    q.append(t.children[i])

            head = head + 1
        return q[0].to_list(tgt_vocab)

    @staticmethod
    def deduplicate_tree(r_list, tgt_vocab):
        q = [Tree.convert_to_tree(r_list, 0, len(r_list), tgt_vocab)]
        head = 0
        while head < len(q):
            t = q[head]
            if len(t.children) > 0:
                k = {}
                for i in range(len(t.children)):
                    if isinstance(t.children[i], Tree):
                        k[t.children[i].to_string().strip()] = i
                    else:
                        k[str(t.children[i]).strip()] = i
                for index in (list(range(len(t.children)))):
                    if index not in ([item[1] for item in k.items()]):
                        t.children.pop(index)
                        t.num_children = t.num_children - 1

            for i in range(len(t.children)):
                if isinstance(t.children[i], Tree):
                    q.append(t.children[i])
            head = head + 1
        return q[0].to_list(tgt_vocab)

    @staticmethod
    def convert_to_tree(r_list, i_left, i_right, tgt_vocab):
        t = Tree()
        level = 0
        left = -1
        for i in range(i_left, i_right):
            if r_list[i] == tgt_vocab.get_symbol_idx('('):
                if level == 0:
                    left = i
                level = level + 1
            elif r_list[i] == tgt_vocab.get_symbol_idx(')'):
                level = level - 1
                if level == 0:
                    if i == left+1:
                        c = r_list[i]
                    else:
                        c = Tree.convert_to_tree(
                            r_list, left + 1, i, tgt_vocab)
                    t.add_child(c)
            elif level == 0:
                t.add_child(r_list[i])
        return t

class VocabForAll():
    def __init__(self, in_word_vocab, out_word_vocab, share_vocab):
        self.in_word_vocab = in_word_vocab
        self.out_word_vocab = out_word_vocab
        self.share_vocab = share_vocab

class Vocab():
    def __init__(self, lower_case=True, pretrained_embedding_fn=None, embedding_dims=300):
        self.symbol2idx = {}
        self.idx2symbol = {}
        self.vocab_size = 0

        self.pad_token = '<PAD>'
        self.start_token = '<START>'
        self.end_token = '<END>'
        self.unk_token = '<UNK>'
        self.non_terminal_token = '<NON-TERMINAL>'

        self.add_symbol(self.pad_token)
        self.add_symbol(self.start_token)
        self.add_symbol(self.end_token)
        self.add_symbol(self.unk_token)
        self.add_symbol(self.non_terminal_token)

        self.lower_case = lower_case
        self.embedding_dims = embedding_dims
        self.embeddings = None

        self.pretrained_embedding_fn = pretrained_embedding_fn

    def add_symbol(self, s):
        if s not in self.symbol2idx:
            self.symbol2idx[s] = self.vocab_size
            self.idx2symbol[self.vocab_size] = s
            self.vocab_size += 1
        return self.symbol2idx[s]

    def get_symbol_idx(self, s):
        if s not in self.symbol2idx:
            return self.get_symbol_idx(self.unk_token)
        else:
            return self.symbol2idx[s]

    def get_idx_symbol(self, idx):
        if idx not in self.idx2symbol:
            return self.unk_token
        return self.idx2symbol[idx]

    def init_from_file(self, fn, min_freq, max_vocab_size):
        print("loading vocabulary file: {}".format(fn))
        with open(fn, "r") as f:
            for line in f:
                l_list = line.strip().split('\t')
                c = int(l_list[1])
                if c >= min_freq:
                    self.add_symbol(l_list[0])
                if self.vocab_size > max_vocab_size:
                    break
        if self.pretrained_embedding_fn is None:
            self.randomize_embeddings(self.embedding_dims)
        else:
            print("loadding pretrained embedding file in {}".format(self.pretrained_embedding_fn))
            self.load_embeddings(self.pretrained_embedding_fn)

    def init_from_list(self, arr, min_freq=1, max_vocab_size=100000):
        for word_, c_ in arr:
            if c_ >= min_freq:
                self.add_symbol(word_)
            if self.vocab_size > max_vocab_size:
                break
        if self.pretrained_embedding_fn is None:
            self.randomize_embeddings(self.embedding_dims)
        else:
            print("loadding pretrained embedding file in {}".format(self.pretrained_embedding_fn))
            self.load_embeddings(self.pretrained_embedding_fn)

    def get_symbol_idx_for_list(self, l):
        r = []
        for i in range(len(l)):
            r.append(self.get_symbol_idx(l[i]))
        return r
    
    def get_idx_symbol_for_list(self, l):
        l = np.array(l)
        r = []
        for i in range(len(l)):
            r.append(self.get_idx_symbol(l[i]))
        return r

    def load_embeddings(self, file_path, scale=0.08, dtype=np.float32):
        """Load pretrained word embeddings for initialization"""
        hit_words = set()
        vocab_size = len(self)
        with open(file_path, 'rb') as f:
            for line in f:
                line = line.split()
                word = line[0].decode('utf-8')
                if self.lower_case:
                    word = word.lower()

                idx = self.get_symbol_idx(word)
                if idx == self.get_symbol_idx(self.unk_token) or idx in hit_words:
                    continue

                vec = np.array(line[1:], dtype=dtype)
                if self.embeddings is None:
                    n_dims = len(vec)
                    self.embeddings = np.array(np.random.uniform(low=-scale, high=scale, size=(vocab_size, n_dims)),
                                               dtype=dtype)
                    self.embeddings[self.get_symbol_idx(self.pad_token)] = np.zeros(n_dims)
                self.embeddings[idx] = vec
                hit_words.add(idx)
        print('Pretrained word embeddings hit ratio: {}'.format(len(hit_words) / len(self.index2word)))

    def randomize_embeddings(self, n_dims, scale=0.08):
        """Use random word embeddings for initialization."""
        vocab_size = self.vocab_size
        shape = (vocab_size, n_dims)
        self.embeddings = np.array(np.random.uniform(low=-scale, high=scale, size=shape), dtype=np.float32)
        self.embeddings[self.get_symbol_idx(self.pad_token)] = np.zeros(n_dims)

    def __getitem__(self, item):
        if type(item) is int:
            return self.get_idx_symbol(item)
        return self.get_symbol_idx(item)

    def __len__(self):
        return self.vocab_size


class DataLoaderForSeqEncoder():
    def __init__(self, data_dir, use_copy, src_vocab_file, tgt_vocab_file, data_file, mode, min_freq, max_vocab_size, batch_size, device):
        self.mode = mode
        self.device = device
        self.data_file = data_file
        self.src_vocab = Vocab()
        self.tgt_vocab = Vocab()
        self.use_copy = use_copy

        if use_copy:
            share_vocab_file_path = self.generate_share_vocab_file(data_dir, src_vocab_file, tgt_vocab_file, min_freq, max_vocab_size)

        print("loading vocabulary file...")
        if not use_copy:
            self.src_vocab.init_from_file(src_vocab_file, min_freq, max_vocab_size)
            self.tgt_vocab.init_from_file(tgt_vocab_file, 0, max_vocab_size)
        else:
            self.src_vocab.init_from_file(share_vocab_file_path, min_freq, max_vocab_size)
            self.tgt_vocab = self.src_vocab
            self.share_vocab = self.src_vocab
            
        print("loading data file...")
        self.data = self._get_seq_tree_pair_data()

        if self.mode == "test":
            assert(batch_size == 1)

        print("padding data with batch_size")
        if len(self.data) % batch_size != 0:
            n = len(self.data)
            for i in range(batch_size - len(self.data) % batch_size):
                self.data.insert(n-i-1, copy.deepcopy(self.data[n-i-1]))

        self.enc_batch_list = []
        self.enc_len_batch_list = []
        self.dec_batch_list = []

        p = 0
        while p + batch_size <= len(self.data):
            max_len = len(self.data[p + batch_size - 1][0])
            m_text = torch.zeros((batch_size, max_len + 2), dtype=torch.long)
            m_text = to_cuda(m_text, device)

            enc_len_list = []
            # add start token
            m_text[:, 0] = self.src_vocab.get_symbol_idx(self.src_vocab.start_token)
            for i in range(batch_size):
                w_list = self.data[p + i][0]
                # print(w_list)
                # reversed order
                for j in range(len(w_list)):
                    m_text[i][j+1] = w_list[len(w_list) - j - 1]
                # add end token
                add_end = 0
                for j in range(len(w_list)+1, max_len+2):
                    if add_end == 0:
                        m_text[i][j] = self.src_vocab.get_symbol_idx(self.src_vocab.end_token)
                        add_end = 1
                    else:
                        m_text[i][j] = self.src_vocab.get_symbol_idx(self.src_vocab.pad_token)
                # print(m_text[i], '\n')
                enc_len_list.append(len(w_list)+2)
            self.enc_batch_list.append(m_text)
            self.enc_len_batch_list.append(enc_len_list)

            tree_batch = []
            for i in range(batch_size):
                tree_batch.append(self.data[p+i][2])
            self.dec_batch_list.append(tree_batch)
            p += batch_size

        self.num_batch = len(self.enc_batch_list)
        assert(len(self.enc_batch_list) == len(self.dec_batch_list))
        print("---Done---")

    def generate_share_vocab_file(self, data_dir, src_vocab_file, tgt_vocab_file, min_freq, max_vocab_size):
        self.src_vocab.init_from_file(src_vocab_file, min_freq, max_vocab_size)
        self.tgt_vocab.init_from_file(tgt_vocab_file, 0, max_vocab_size)

        share_vocab = {}
        num = np.max([self.src_vocab.vocab_size, self.tgt_vocab.vocab_size]) + min_freq + 1

        for word, freq in self.src_vocab.symbol2idx.items():
            freq = num-freq
            if word not in share_vocab.keys():
                share_vocab[word] = freq
            else:
                share_vocab[word] += freq

        for word, freq in self.tgt_vocab.symbol2idx.items():
            freq = num-freq
            if word not in share_vocab.keys():
                share_vocab[word] = freq
            else:
                share_vocab[word] += freq

        share_vocab = sorted(share_vocab.items(), key=lambda d:d[1], reverse = True)[5:]
        with open(data_dir + 'share_vocab.txt','w') as f:
            for k, v in share_vocab:
                f.write(k+'\t'+str(v)+'\n')
        return (data_dir + 'share_vocab.txt')


    def _get_seq_tree_pair_data(self):
        data = []
        with open(self.data_file, "r") as f:
            for line in f:
                l_list = line.strip().split("\t")
                w_list = self.src_vocab.get_symbol_idx_for_list(
                    l_list[0].strip().split(' '))
                r_list = self.tgt_vocab.get_symbol_idx_for_list(
                    l_list[1].strip().split(' '))
                cur_tree = Tree.convert_to_tree(
                    r_list, 0, len(r_list), self.tgt_vocab)
                data.append((w_list, r_list, cur_tree))
        return data

    def random_batch(self):
        p = randint(0, self.num_batch-1)
        return self.enc_batch_list[p], self.enc_len_batch_list[p], self.dec_batch_list[p]

    def all_batch(self):
        r = []
        for p in range(self.num_batch):
            r.append([self.enc_batch_list[p],
                      self.enc_len_batch_list[p], self.dec_batch_list[p]])
        return r


class DataLoaderForGraphEncoder():
    def __init__(self, use_copy, use_share_vocab, data, dataset, mode, batch_size, device):
        self.mode = mode
        self.device = device
        self.use_copy = use_copy
        self.batch_size = batch_size
        self.src_vocab = dataset.src_vocab_model
        self.tgt_vocab = dataset.tgt_vocab_model
        if use_share_vocab and self.src_vocab.vocab_size == self.tgt_vocab.vocab_size:
            self.share_vocab = self.src_vocab
        
        self.data = [(item.graph, item.output_text, item.output_tree) for item in data]

        if self.mode == "test":
            assert(batch_size == 1)

        if self.mode == "train" and len(self.data) % batch_size != 0:
            n = len(self.data)
            for i in range(batch_size - len(self.data) % batch_size):
                self.data.insert(n-i-1, copy.deepcopy(self.data[n-i-1]))

        self.enc_batch_list = []
        self.enc_len_batch_list = []
        self.original_target_tree_list = []
        self.dec_batch_list = []

        p = 0
        while p + batch_size <= len(self.data):
            vectorized_batch_graph = []
            tree_batch = []
            original_target_tree = []
            enc_len_batch = 0

            for i in range(batch_size):
                graph_item = self.data[p + i][0]
                vectorized_batch_graph.append(graph_item)
                enc_len_batch += graph_item.get_node_num()
                original_target_tree.append(self.data[p+i][1])
                tree_batch.append(self.data[p+i][2])
            
            self.enc_batch_list.append(vectorized_batch_graph)
            self.enc_len_batch_list.append(enc_len_batch)
            self.original_target_tree_list.append(original_target_tree)
            self.dec_batch_list.append(tree_batch)
            p += batch_size

        self.num_batch = len(self.enc_batch_list)
        assert(len(self.enc_batch_list) == len(self.dec_batch_list))


    def random_batch(self):
        p = randint(0, self.num_batch-1)
        return {'encoder_graph_batch': self.enc_batch_list[p],
                'encoder_graph_node_number_batch': self.enc_len_batch_list[p],
                'decoder_tree_batch': self.dec_batch_list[p],
                'decoder_output_text_batch': self.original_target_tree_list[p]
                }
                
    def all_batch(self):
        r = []
        for p in range(self.num_batch):
            r.append([self.enc_batch_list[p],
                      self.enc_len_batch_list[p], self.dec_batch_list[p]], self.original_target_tree_list[p])
        return r


def to_cuda(x, device=None):
    if device:
        x = x.to(device)
    return x



import operator
from queue import PriorityQueue

import torch
import torch.nn as nn

from graph4nlp.pytorch.data.data import GraphData
from graph4nlp.pytorch.modules.prediction.generation.base import DecoderBase
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
from copy import deepcopy


class StrategyBase(nn.Module):
    def __init__(self):
        super(StrategyBase, self).__init__()

    def generate(self, **kwargs):
        raise NotImplementedError()


class BeamSearchNode(object):
    def __init__(self, hiddenstate, enc_attn_weights_average, previousNode, wordId, logProb, length):
        self.h = hiddenstate
        self.enc_attn_weights_average = enc_attn_weights_average
        self.prevNode = previousNode
        self.wordid = wordId
        self.logp = logProb
        self.leng = length

    def eval(self, alpha=1.0):
        reward = 0
        # Add here a function for shaping a reward

        return self.logp / float(self.leng - 1 + 1e-6) + alpha * reward


class Hypothesis(object):
    def __init__(self, tokens, log_probs, dec_state, num_non_words, enc_attn_weights, use_coverage):
        self.tokens = tokens
        self.log_probs = log_probs
        self.dec_state = dec_state
        self.num_non_words = num_non_words
        self.enc_attn_weights = enc_attn_weights
        self.use_coverage = use_coverage
        if not self.use_coverage:
            self.enc_attn_weights = []

    def __repr__(self):
        return repr(self.tokens)

    def __len__(self):
        return len(self.tokens) - self.num_non_words

    @property
    def avg_log_prob(self):
        return sum(self.log_probs) / len(self.log_probs)

    def create_next(self, token, log_prob, dec_state, non_word, add_enc_attn_weights):
        if self.use_coverage:
            assert len(add_enc_attn_weights.shape) == 3
            assert add_enc_attn_weights.shape[0] == 1
            assert add_enc_attn_weights.shape[1] == 1
            enc_attn_weights_processed = [t.clone() for t in self.enc_attn_weights]
            enc_attn_weights_processed.append(add_enc_attn_weights)
        else:
            enc_attn_weights_processed = []
        return Hypothesis(tokens=self.tokens + [token], log_probs=self.log_probs + [log_prob],
                          dec_state=dec_state,
                          num_non_words=self.num_non_words + 1 if non_word else self.num_non_words,
                          enc_attn_weights=enc_attn_weights_processed, use_coverage=self.use_coverage)



class DecoderStrategy(StrategyBase):
    def __init__(self, beam_size, vocab, decoder: DecoderBase, rnn_type, use_copy=False, use_coverage=False,
                 max_decoder_step=50):
        super(DecoderStrategy, self).__init__()
        self.rnn_type = rnn_type
        self.beam_size = beam_size
        self.decoder = decoder
        self.vocab = vocab
        self.use_copy = use_copy
        self.use_coverage = use_coverage
        self.max_decoder_step = max_decoder_step

    def generate(self, graph_list: list, oov_dict=None, topk=1):
        """
            Generate sequences using beam search.
        Parameters
        ----------
        graph_list: list[GraphData]
        oov_dict: Vocab
        topk: int, default=1

        Returns
        -------
        prediction: list
        """
        params = self.decoder.extract_params(graph_list)
        params['tgt_seq'] = None
        params['beam_size'] = self.beam_size
        params['topk'] = topk
        params['oov_dict'] = oov_dict
        return self._beam_search(**params)

    def _beam_search(self, graph_node_embedding, graph_node_mask=None, rnn_node_embedding=None,
                     graph_level_embedding=None,
                     graph_edge_embedding=None, graph_edge_mask=None, tgt_seq=None, src_seq=None, oov_dict=None,
                     beam_size=4, topk=1):

        assert 0 < topk <= beam_size
        len_in_words = False

        min_out_len = 1
        max_out_len = self.max_decoder_step + 1
        batch_size = graph_node_embedding.shape[0]

        decoder_state = self.decoder.get_decoder_init_state(rnn_type=self.rnn_type, batch_size=batch_size,
                                                            content=graph_level_embedding)

        batch_results = []
        for batch_idx in range(batch_size):
            if self.rnn_type == "lstm":
                single_decoder_state = (
                    decoder_state[0][:, batch_idx, :].unsqueeze(1), decoder_state[1][:, batch_idx, :].unsqueeze(1))
            elif self.rnn_type == "gru":
                single_decoder_state = decoder_state[:, batch_idx, :].unsqueeze(1)
            else:
                raise NotImplementedError("RNN Type {} is not implemented, expected in ``lstm`` or ``gru``"
                                          .format(self.rnn_type))

            single_graph_node_embedding = graph_node_embedding[batch_idx, :, :].unsqueeze(0).expand(beam_size, -1, -1).contiguous()
            single_graph_node_mask = graph_node_mask[batch_idx, :].unsqueeze(0).expand(beam_size, -1, -1).contiguous() if graph_node_mask is not None else None
            single_rnn_node_embedding = rnn_node_embedding[batch_idx, :, :].unsqueeze(
                0).expand(beam_size, -1, -1).contiguous() if rnn_node_embedding is not None else None

            step = 0
            results, backup_results = [], []

            hypos = [Hypothesis([self.vocab.SOS], [], single_decoder_state, 1, [], use_coverage=self.use_coverage)]

            while len(hypos) > 0 and step <= self.max_decoder_step:
                n_hypos = len(hypos)
                if n_hypos < beam_size:
                    hypos.extend(deepcopy(hypos[-1]) for _ in range(beam_size - n_hypos)) # check deep copy

                decoder_input = torch.tensor([h.tokens[-1] for h in hypos]).to(graph_node_embedding.device)

                if self.rnn_type == "lstm":
                    single_decoder_state = (
                        torch.cat([h.dec_state[0] for h in hypos], 1), torch.cat([h.dec_state[1] for h in hypos], 1))
                elif self.rnn_type == "gru":
                    single_decoder_state = torch.cat([h.dec_state for h in hypos], 1)
                else:
                    raise NotImplementedError("RNN Type {} is not implemented, expected in ``lstm`` or ``gru``"
                                              .format(self.rnn_type))

                decoder_input = self.decoder._filter_oov(decoder_input)

                enc_tmp = []
                coverage_input = []
                if step > 0  and self.use_coverage:
                    for ii in range(step):
                        coverage_input.append(torch.cat([h.enc_attn_weights[ii] for h in hypos], dim=1))

                decoder_output, single_decoder_state, dec_attn_scores, _ = \
                    self.decoder.decode_step(decoder_input=decoder_input, rnn_state=single_decoder_state,
                                             dec_input_mask=single_graph_node_mask,
                                             encoder_out=single_graph_node_embedding, rnn_emb=single_rnn_node_embedding,
                                             enc_attn_weights_average=coverage_input,
                                             src_seq=src_seq[batch_idx, :].unsqueeze(0) if self.use_copy else None,
                                             oov_dict=oov_dict)
                decoder_output = torch.log(decoder_output + 1e-31)
                top_v, top_i = decoder_output.data.topk(beam_size)
                # print(top_v.shape, decoder_output.shape, dec_attn_scores.shape, "-----")

                new_hypos = []
                for in_idx in range(n_hypos):
                    for out_idx in range(beam_size):
                        new_tok = top_i[in_idx][out_idx].item()
                        new_prob = top_v[in_idx][out_idx].item()
                        new_enc_attn_weights = dec_attn_scores[in_idx, :].unsqueeze(0).unsqueeze(0)

                        non_word = new_tok == self.vocab.EOS  # only SOS & EOS don't count

                        if self.rnn_type == "lstm":
                            tmp_decoder_state = (
                                single_decoder_state[0][:, in_idx, :].unsqueeze(1),
                                single_decoder_state[1][:, in_idx, :].unsqueeze(1))
                        elif self.rnn_type == "gru":
                            tmp_decoder_state = single_decoder_state[:, in_idx, :].unsqueeze(1)
                        else:
                            raise NotImplementedError("RNN Type {} is not implemented, expected in ``lstm`` or ``gru``"
                                                      .format(self.rnn_type))

                        new_hypo = hypos[in_idx].create_next(token=new_tok, log_prob=new_prob,
                                                             dec_state=tmp_decoder_state, non_word=non_word,
                                                             add_enc_attn_weights=new_enc_attn_weights)
                        new_hypos.append(new_hypo)

                # Block sequences with repeated ngrams
                # block_ngram_repeats(new_hypos, block_ngram_repeat)

                # process the new hypotheses
                new_hypos = sorted(new_hypos, key=lambda h: -h.avg_log_prob)[:beam_size]
                hypos = []
                new_complete_results, new_incomplete_results = [], []
                for nh in new_hypos:
                    length = len(nh)  # Does not count SOS and EOS
                    if nh.tokens[-1] == self.vocab.EOS:  # a complete hypothesis
                        if len(new_complete_results) < beam_size and min_out_len <= length <= max_out_len:
                            new_complete_results.append(nh)
                    elif len(hypos) < beam_size and length < max_out_len:  # an incomplete hypothesis
                        hypos.append(nh)
                    elif length == max_out_len and len(new_incomplete_results) < beam_size:
                        new_incomplete_results.append(nh)
                if new_complete_results:
                    results.extend(new_complete_results)
                elif new_incomplete_results:
                    backup_results.extend(new_incomplete_results)
                step += 1
            if not results:  # if no sequence ends with EOS within desired length, fallback to sequences
                results = backup_results  # that are "truncated" at the end to max_out_len
            batch_results.append(sorted(results, key=lambda h: -h.avg_log_prob)[:beam_size])
        ret = torch.zeros(batch_size, topk, self.max_decoder_step).long()
        for sent_id, each in enumerate(batch_results):
            for i in range(topk):
                ids = torch.Tensor(each[i].tokens[1:])[:self.max_decoder_step]
                if ids.shape[0] < self.max_decoder_step:
                    pad = torch.zeros(self.max_decoder_step - ids.shape[0])
                    ids = torch.cat((ids, pad), dim=0)
                ret[sent_id, i, :] = ids
        return ret

    def beam_search_for_tree_decoding(self, decoder_initial_state,
                                        decoder_initial_input,
                                        parent_state,
                                        graph_node_embedding,
                                        rnn_node_embedding=None,
                                        src_seq=None,
                                        oov_dict=None,
                                        sibling_state=None,
                                        device=None,
                                        topk=1,
                                        enc_batch=None):

        decoded_results = []

        decoder_hidden = decoder_initial_state
        decoder_input = decoder_initial_input
        form_manager = self.vocab

        # Number of sentence to generate
        endnodes = []
        number_required = min((topk + 1), topk - len(endnodes))

        # starting node -  hidden vector, previous node, word id, logp, length
        node = BeamSearchNode(decoder_hidden, [], None, decoder_input, 0, 1)
        nodes = PriorityQueue()

        # start the queue
        nodes.put((-node.eval(), torch.randn(1).item(), node))
        qsize = 1

        # start beam search
        while True:
            if qsize > self.max_decoder_step: break

            # fetch the best node
            score, _nounce, n = nodes.get()
            decoder_input = n.wordid
            decoder_hidden = n.h

            if n.wordid.item() == form_manager.get_symbol_idx(form_manager.end_token) and n.prevNode != None:
                endnodes.append((score, _nounce, n))
                # if we reached maximum # of sentences required
                if len(endnodes) >= number_required:
                    break
                else:
                    continue
                
            # decode for one step using decoder
            prediction, decoder_hidden, _ = self.decoder.decode_step(tgt_batch_size=1, 
                                                        dec_single_input=decoder_input,
                                                        dec_single_state=decoder_hidden,
                                                        memory=graph_node_embedding,
                                                        parent_state=parent_state,
                                                        oov_dict=oov_dict,
                                                        enc_batch=enc_batch)

            # PUT HERE REAL BEAM SEARCH OF TOP
            log_prob, indexes = torch.topk(prediction, self.beam_size)
            nextnodes = []

            for new_k in range(self.beam_size):
                decoded_t = torch.tensor([indexes[0][new_k]], dtype=torch.long, device=device)

                log_p = log_prob[0][new_k].item()

                node = BeamSearchNode(decoder_hidden, [], n, decoded_t, n.logp + log_p, n.leng + 1)
                score = -node.eval()
                nextnodes.append((score, torch.randn(1).item(), node))

            # put them into queue
            for i in range(len(nextnodes)):
                score, _nounce, nn = nextnodes[i]
                nodes.put((score, _nounce, nn))
                # increase qsize
            qsize += len(nextnodes) - 1

        # choose nbest paths, back trace them
        if len(endnodes) == 0:
            endnodes = [nodes.get() for _ in range(topk)]

        utterances = []
        for score, _, n in sorted(endnodes, key=operator.itemgetter(0)):
            utterance = []
            utterance.append(n)
            # back trace
            while n.prevNode != None:
                n = n.prevNode
                utterance.append(n)

            utterance = utterance[::-1]
            utterances.append(utterance)

        decoded_results.append(utterances)
        assert(len(decoded_results) == 1 and len(utterances) == topk)
        return decoded_results


import numpy as np
import torch


class EarlyStopping:
    def __init__(self, save_model_path, patience=10):
        self.patience = patience
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.save_model_path = save_model_path

    def step(self, acc, model):
        score = acc
        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(model)
        elif score < self.best_score:
            self.counter += 1
            print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(model)
            self.counter = 0

        return self.early_stop

    def save_checkpoint(self, model):
        '''Saves model when validation loss decrease.'''
        torch.save(model.state_dict(), self.save_model_path)
        print('Saved model to {}'.format(self.save_model_path))

    def load_checkpoint(self, model):
        model.load_state_dict(torch.load(self.save_model_path))

        return model



from ....modules.utils.config_utils import get_yaml_config
import os

str2yaml = {"dependency": "dependency.yaml", "constituency": "constituency.yaml", "ie": "ie_graph_construction.yaml",
            "node_emb": "node_emb.yaml", "node_emb_refined": "node_emb_refine.yaml"}


def get_graph_construction_args(graph_construction_name):
    """
        It will build the template for ``GraphConstructionBase`` model.
    Parameters
    ----------
    graph_construction_name: str
        The graph construction method name. Expected in ["dependency", "constituency", "ie", "node_emb", "node_emb_refined"].
        If it can't find the ``graph_construction_name``, it will return ``{}``.
    Returns
    -------
    template_dict: dict
        The template dict.
        The structure is shown as follows:
        {
            graph_construction_share: {graph_type: 'constituency', root_dir: null, ...},
            graph_construction_private: {merge_strategy: 'tailhead', sequential_link: True},
            node_embedding: {
                                input_size: 300,
                                ...,
                                embedding_style: {single_token_item: True, ...}
                            }
        }
        The ``graph_construction_share`` contains the parameters shared by all ``GraphConstructionBase`` models.
        The ``graph_construction_private`` contains the parameters specifically in each graph_construction methods.
        The ``node_embedding`` contains all the parameters for node embedding.
            Specifically, it contains ``embedding_style`` which for embedding style.
    """
    if graph_construction_name in str2yaml.keys():
        yaml_name = str2yaml[graph_construction_name]
        path = os.path.join("graph4nlp/pytorch/modules/config/graph_construction", yaml_name)
        config = get_yaml_config(path)
        return config
    else:
        return {}


__all__ = ["get_graph_construction_args"]



import argparse

from graph4nlp.pytorch.modules.config import get_basic_args
from graph4nlp.pytorch.modules.utils.config_utils import update_values, get_yaml_config


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset_yaml", type=str,
                        default="examples/pytorch/semantic_parsing/graph2seq/config/new_dependency_gat_bi_sep.yaml", help="")
                        # default="examples/pytorch/semantic_parsing/graph2seq/config/new_dynamic_refine.yaml", help="")
                        # default = "examples/pytorch/semantic_parsing/graph2seq/config/new_dynamic.yaml", help = "")

                        # default="examples/pytorch/semantic_parsing/graph2seq/config/new_constituency.yaml", help="")

    parser.add_argument('--word-emb-size', type=int, default=300, help='')
    parser.add_argument("--log-file", type=str, default="examples/pytorch/semantic_parsing/graph2seq/log/ggnn.txt")
    parser.add_argument("--checkpoint-save-path", type=str, default="examples/pytorch/semantic_parsing/graph2seq/save")

    parser.add_argument('--learning-rate', type=float, default=1e-3, help='')
    parser.add_argument("--loss-display-step", type=int, default=3, help=' ')
    parser.add_argument("--eval-display-number", type=int, default=3, help="")
    parser.add_argument("--lr-start-decay-epoch", type=int, default=20, help="")
    parser.add_argument("--lr-decay-rate", type=float, default=0.9)
    parser.add_argument("--lr-decay-per-epoch", type=int, default=5, help="")
    parser.add_argument("--min-lr", type=float, default=1e-3, help="")
    parser.add_argument("--use-gpu", type=float, default=1, help="0 for don't use cuda, 1 for using cuda")
    parser.add_argument("--gpu", type=int, default=0, help="gpu id")
    parser.add_argument("--seed", type=int, default=1236, help="")

    # dataset config
    parser.add_argument("--batch_size", type=int, default=24, help="the size of one mini-batch")
    parser.add_argument("--share-vocab", type=bool, default=True, help="whether to share vocab")
    parser.add_argument("--val_split_ratio", type=float, default=0, help="")
    parser.add_argument("--pretrained_word_emb_file", type=str,
                        default='/home/shiina/shiina/lib/graph4nlp/.vector_cache/glove.6B.300d.txt', help="")

    parser.add_argument("--beam-size", type=int, default=4, help="the beam size of beam search")

    cfg = parser.parse_args()

    our_args = get_yaml_config(cfg.dataset_yaml)
    template = get_basic_args(graph_construction_name=our_args["graph_construction_name"],
                              graph_embedding_name=our_args["graph_embedding_name"],
                              decoder_name=our_args["decoder_name"])
    update_values(to_args=template, from_args_list=[our_args, vars(cfg)])
    return template



import torch
import torch.nn as nn


class Attention(nn.Module):
    def __init__(self, query_size, memory_size, hidden_size, has_bias=False, dropout=0.2, attention_funtion="mlp"):
        super(Attention, self).__init__()
        self.query_size = query_size
        self.memory_size = memory_size
        self.hidden_size = hidden_size
        self.attn_type = attention_funtion
        self.dropout = nn.Dropout(dropout)
        assert self.attn_type in ["mlp", "general", "dot"]
        if self.attn_type == "mlp":
            self.query_in = nn.Linear(self.query_size, self.hidden_size, bias=True if has_bias else False)
            self.memory_in = nn.Linear(self.memory_size, self.hidden_size, bias=False)
        elif self.attn_type == "dot":
            if memory_size != query_size:
                raise ValueError("Parameter \"memory_size\" must be equal to \"query_size\"")
        elif self.attn_type == "general":
            self.query2memory = nn.Linear(query_size, memory_size, bias=False)
        self.out = nn.Linear(self.hidden_size, 1, bias=False)
        self.inf = 1e8

    def forward(self, query, memory, memory_mask=None, coverage=None):
        """
            Attention function
        Parameters
        ----------
        query: torch.Tensor, shape=[B, D]
        memory: torch.Tensor, shape=[B, N, D]
        memory_mask: torch.Tensor, shape=[B, N]
        coverage: torch.Tensor, shape=[B, N, D]

        Returns
        -------
        attn_results: torch.Tensor, shape=[B, D]
        attn_scores: torch.Tensor, shape=[B, N]
        """
        assert len(query.shape) == 2
        assert len(memory.shape) == 3
        assert query.shape[0] == memory.shape[0]

        aligns = self._calculate_aligns(query, memory, coverage=coverage)

        if memory_mask is not None:
            aligns = aligns.masked_fill(memory_mask == 0, -self.inf)
        scores = torch.softmax(aligns, dim=-1)
        scores = self.dropout(scores)
        ret = torch.bmm(scores.unsqueeze(1), memory)
        return ret.squeeze(1), scores

    def _calculate_aligns(self, src, tgt, coverage=None):
        """
            Attention score calculation.
        Parameters
        ----------
        src: torch.Tensor, shape=[B, D]
        tgt: torch.Tensor, shape=[B, N, D]
        coverage: torch.Tensor, shape=[B, N, D]
        Returns
        -------
        aligns: torch.Tensor, shape=[B, N]
        """
        if self.attn_type == "mlp":
            src = self.query_in(src)
            tgt = self.memory_in(tgt)
            aligns = src.unsqueeze(1) + tgt
            if coverage is not None:
                assert len(coverage.shape) == 3
                assert coverage.shape[-1] == self.hidden_size
                aligns += coverage
            aligns = torch.tanh(aligns)
            aligns = self.out(aligns)
            aligns = aligns.view(aligns.shape[0], aligns.shape[1])
        elif self.attn_type == "general" or self.attn_type == "dot":
            if coverage:
                assert tgt.shape == coverage.shape
                tgt = tgt + coverage
                tgt = torch.tanh(tgt)

            if self.attn_type == "general":
                src = self.query2memory(src)
            aligns = torch.bmm(src.unsqueeze(1), tgt.transpose(1, 2))
            aligns = aligns.squeeze(1).contiguous()
        else:
            raise NotImplementedError()
        return aligns


import torch
from .StackedElementProdLayer import StackedElementProdLayer
import torch

from .StackedElementProdLayer import StackedElementProdLayer


class StackedElementProd(StackedElementProdLayer):
    r"""Specific class for link prediction task.

    Parameters
    ----------

    input_size : int 
                 The length of input node embeddings
    num_class : int 
               The number of node catrgoriey for classification
    num_channel: int
               The number of channels for node embeddings to be used for link prediction                 
    hidden_size : list of int type values
                  Example for two layers's FeedforwardNN: [50, 20]        

    """

    def __init__(self, input_size, hidden_size, num_class):
        super(StackedElementProd, self).__init__()
        
        self.num_channel=num_class
        self.classifier=StackedElementProdLayer(input_size, num_class, self.num_channel, hidden_size)

    def forward(self, input_graph):
        r"""
        Forward functions to compute the logits tensor for link prediction.
    
      
        Parameters
        ----------
    
        input graph : GraphData
                     The tensors stored in the node feature field named as  
                     "node_emb_"+num_channel (start from "node_emb_0") 
                     in the input_graph are used  for link prediction.
                     

    
        Returns 
        ---------
        
        output_graph : GraphData
                      The computed logit tensor for each pair of nodes in the graph are stored
                      in the node feature field named "edge_logits".
                      logit tensor shape is: [num_class] 
        """
        # get the nod embedding from the graph
        node_emb_list = []
        for channel_idx in range(self.num_channel):
           node_emb_list.append(input_graph.node_features['node_emb_'+str(channel_idx)])
        
        #add the edges and edge prediction logits into the graph
        num_node=node_emb_list[0].shape[1]
        node_idx_list=[idx for idx in range(num_node)]
        src_idx=torch.tensor(node_idx_list).view(-1,1).repeat(1,num_node).view(-1)
        dst_idx=torch.tensor(node_idx_list).view(1,-1).repeat(num_node,1).view(-1)
        input_graph.add_edges(src_idx,dst_idx)
        input_graph.edge_features['logits']=self.classifier(node_emb_list)     
        
        return input_graph



import time
import numpy as np
import torch
import dgl

from ...data.data import GraphData
from ...modules.graph_construction.embedding_construction import EmbeddingConstruction
from ...modules.utils.vocab_utils import VocabModel
from ...modules.utils.padding_utils import pad_2d_vals, pad_2d_vals_no_size


if __name__ == '__main__':
    raw_text_data = [['Google News is a news aggregator service developed by Google.',
                    'It presents a continuous flow of articles organized from thousands of publishers and magazines.'],
                    ['Google News is available as an app on Android, iOS, and the Web.',
                     'Google released a beta version in September 2002 and the official app in January 2006.']]
    vocab_model = VocabModel(raw_text_data, max_word_vocab_size=None,
                            min_word_vocab_freq=1,
                            pretrained_word_emb_file=None,
                            word_emb_size=300)

    batch_size = 64
    hidden_size = 128
    num_nodes = list(range(3, 3 + batch_size))
    np.random.shuffle(num_nodes)

    # build graph
    max_node_len = 0
    graph_list = []
    node_size = []
    for i in range(batch_size):
        graph = GraphData()
        graph.add_nodes(num_nodes[i])
        for j in range(num_nodes[i]):
            tokens = np.random.choice(list(vocab_model.word_vocab.word2index.keys()), np.random.choice(range(1, 7)))
            graph.node_attributes[j]['token_idx'] = vocab_model.word_vocab.to_index_sequence_for_list(tokens)
            node_size.append(len(graph.node_attributes[j]['token_idx']))
            max_node_len = max(node_size[-1], max_node_len)

        graph_list.append(graph)

    for graph in graph_list:
        tmp_token_idx = [graph.node_attributes[j]['token_idx'] for j in range(graph.get_node_num())]
        tmp_token_idx = torch.LongTensor((pad_2d_vals(tmp_token_idx, len(tmp_token_idx), max_node_len)))
        graph.node_features['token_idx'] = tmp_token_idx

    graph_list = [graph.to_dgl() for graph in graph_list]
    bg = dgl.batch(graph_list, edge_attrs=None)
    node_size = torch.LongTensor(node_size)
    num_nodes = torch.LongTensor(num_nodes)

    emb_constructor = EmbeddingConstruction(vocab_model.word_vocab, 'w2v', 'bilstm', 'bilstm', hidden_size, device=None)
    t0 = time.time()
    node_feat = emb_constructor(bg.ndata['token_idx'], node_size, num_nodes)
    print('runtime: {}'.format(time.time() - t0))
    print('mean', node_feat.mean())

    print('emb: {}'.format(node_feat.shape))



import random

import torch

from ...data.data import GraphData
from datetime import datetime
import random

# Test hyperparameters
num_nodes = 100
num_edges = 50

src_list, tgt_list = [], []
for i in range(num_edges):
    src = random.randint(a=0, b=num_nodes - 1)
    tgt = random.randint(a=0, b=num_nodes - 1)
    while tgt == src:
        tgt = random.randint(a=0, b=num_nodes - 1)
    src_list.append(src)
    tgt_list.append(tgt)

# Test build graph and add nodes
g = GraphData()
g.add_nodes(num_nodes)

# Add node attributes
for i in range(num_nodes):
    g.node_attributes[i]['name'] = '*NAMEFIX_{}'.format(i)
    g.node_attributes[i]['value'] = i

# Test add edges
g.add_edges(src_list, tgt_list)

# Test modify and access node features and attributes
emb_layer = torch.nn.Embedding(100000, 10)
g.node_features['node_feat'] = torch.zeros((num_nodes, 10))
adam = torch.optim.Adam(emb_layer.parameters())


# Do some computation here
for i in range(100):
    for i in range(num_nodes):
        g.node_features['node_feat'][i] = emb_layer(torch.tensor([i], dtype=torch.long))

    t0 = datetime.now()
    dgl_g = g.to_dgl()
    print('Time elapsed for convertion = {}.'.format(datetime.now() - t0))

    loss = -torch.nn.functional.cosine_similarity(dgl_g.ndata['node_feat'].sum(dim=-1), torch.ones(num_nodes), dim=0)
    adam.zero_grad()
    loss.backward(retain_graph=True)
    adam.step()
    print('loss = {}'.format(loss))
pass

# # User Code Proposal 1
# graph_data = graph4nlp.graph_construction.CTGraph(raw_data)
# new_graph_data = graph4nlp.graph_construction.GAT(graph_data)
# result = graph4nlp.prediction.NodeClassfier(new_graph_data)
#
# # User Code Proposal 2
# graph_data = graph4nlp.graph_construction.CTGraph(raw_data)
# dgl_graphs = graph_data.to_dgl()
# features = dgl_graphs.ndata['node_feat']  # The key here ('node_feat') can be anything
# new_feats = graph4nlp.graph_construction.GAT(dgl_graphs, features)
# graph_data.node_features['node_emb'] = new_feats
#
#
# # Library code
# class GAT(SomeBase):
#     def __init__(self, num_layers):
#         self.gat_layers = nn.ModuleList
#
#     def forward(self, graph_data: GraphData):
#         dgl_g = graph_data.to_dgl()
#         features = dgl_g.ndata['node_feat']  # 'node_emb' is the reserved key
#         new_features = self.gat_layers(dgl_g, features)
#         graph_data.node_features['node_emb'] = new_features
#         return graph_data
#
#
# class NodeClassifier(AnotherBase):
#     def __init__(self):
#         self.node_classification_layer = NodeClassificationLayer()
#
#     def forward(self, graph_data: GraphData):
#         node_emb = graph_data.node_features['node_emb']
#         return self.node_classification_layer(node_emb)
#
#
# # User implement another model
# class MyGAT(SomeBase):
#     def forward(self, graph_data: GraphData):
#         dgl_g = graph_data.to_dgl()
#         # Extract features here
#         features_1 = dgl_g.ndata['my_feat']
#         features_2 = dgl_g.ndata['another_feat']
#         new_features_1, new_features_2 = MyGATLayer(dgl_g, features_1, features_2)
#         graph_data.node_features['my_node_emb'] = new_features_1
#         graph_data.node_features['node_emb_2'] = new_features_2
#
#
# pass



"""
graph4nlp.datasets module contains various common datasets implemented based on graph4nlp.data.dataset.
"""

from .cnn import CNNDataset
from .geo import GeoDatasetForTree
from .jobs import JobsDataset, JobsDatasetForTree
from .kinship import KinshipDataset
from .mawps import MawpsDatasetForTree
from .squad import SQuADDataset
from .trec import TrecDataset


import os

# os.environ["CUDA_VISIBLE_DEVICES"] = "2"

# os.environ['CUDA_LAUNCH_BLOCKING'] = "5"
from graph4nlp.pytorch.datasets.jobs import JobsDataset
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.constituency_graph_construction import \
    ConstituencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_graph_construction import \
    NodeEmbeddingBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_refined_graph_construction import \
    NodeEmbeddingBasedRefinedGraphConstruction

import numpy as np
import torch
from torch.utils.data import DataLoader
import torch.optim as optim

from .args import get_args
from .evaluation import ExpressionAccuracy
from .utils import get_log, wordid2str
from .build_model import get_model
from graph4nlp.pytorch.modules.loss.seq_generation_loss import Graph2SeqLoss
from graph4nlp.pytorch.modules.utils.copy_utils import prepare_ext_vocab


class Jobs:
    def __init__(self, opt):
        super(Jobs, self).__init__()
        self.opt = opt
        self.use_copy = self.opt["decoder_args"]["rnn_decoder_share"]["use_copy"]
        self.use_coverage = self.opt["decoder_args"]["rnn_decoder_share"]["use_coverage"]
        self._build_device(self.opt)
        self._build_logger(self.opt["log_file"])
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()
        self._build_loss_function()

    def _build_device(self, opt):
        seed = opt["seed"]
        np.random.seed(seed)
        if opt["use_gpu"] != 0 and torch.cuda.is_available():
            print('[ Using CUDA ]')
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            from torch.backends import cudnn
            cudnn.benchmark = True
            device = torch.device('cuda' if opt["gpu"] < 0 else 'cuda:%d' % opt["gpu"])
        else:
            print('[ Using CPU ]')
            device = torch.device('cpu')
        self.device = device

    def _build_logger(self, log_file):
        self.logger = get_log(log_file)

    def _build_dataloader(self):
        if self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "dependency":
            topology_builder = DependencyBasedGraphConstruction
            graph_type = 'static'
            dynamic_init_topology_builder = None
        elif self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "constituency":
            topology_builder = ConstituencyBasedGraphConstruction
            graph_type = 'static'
            dynamic_init_topology_builder = None
        elif self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "node_emb":
            topology_builder = NodeEmbeddingBasedGraphConstruction
            graph_type = 'dynamic'
            dynamic_init_topology_builder = None
        elif self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "node_emb_refined":
            topology_builder = NodeEmbeddingBasedRefinedGraphConstruction
            graph_type = 'dynamic'
            dynamic_init_graph_type = self.opt["graph_construction_args"]["graph_construction_private"][
                "dynamic_init_graph_type"]
            if dynamic_init_graph_type is None or dynamic_init_graph_type == 'line':
                dynamic_init_topology_builder = None
            elif dynamic_init_graph_type == 'dependency':
                dynamic_init_topology_builder = DependencyBasedGraphConstruction
            elif dynamic_init_graph_type == 'constituency':
                dynamic_init_topology_builder = ConstituencyBasedGraphConstruction
            else:
                # dynamic_init_topology_builder
                raise RuntimeError('Define your own dynamic_init_topology_builder')
        else:
            raise NotImplementedError("Define your topology builder.")

        dataset = JobsDataset(root_dir=self.opt["graph_construction_args"]["graph_construction_share"]["root_dir"],
                              pretrained_word_emb_file=self.opt["pretrained_word_emb_file"],
                              val_split_ratio=self.opt["val_split_ratio"],
                              merge_strategy=self.opt["graph_construction_args"]["graph_construction_private"][
                                  "merge_strategy"],
                              edge_strategy=self.opt["graph_construction_args"]["graph_construction_private"][
                                  "edge_strategy"],
                              seed=self.opt["seed"],
                              word_emb_size=self.opt["word_emb_size"], share_vocab=self.opt["share_vocab"],
                              graph_type=graph_type,
                              topology_builder=topology_builder,
                              topology_subdir=self.opt["graph_construction_args"]["graph_construction_share"][
                                  "topology_subdir"],
                              dynamic_graph_type=self.opt["graph_construction_args"]["graph_construction_share"][
                                  "graph_type"],
                              dynamic_init_topology_builder=dynamic_init_topology_builder,
                              dynamic_init_topology_aux_args=None)

        self.train_dataloader = DataLoader(dataset.train, batch_size=self.opt["batch_size"], shuffle=True,
                                           num_workers=1,
                                           collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset.test, batch_size=self.opt["batch_size"], shuffle=False, num_workers=1,
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model

    def _build_model(self):

        self.model = get_model(self.opt, vocab_model=self.vocab, device=self.device).to(self.device)

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=self.opt["learning_rate"])

    def _build_evaluation(self):
        self.metrics = [ExpressionAccuracy()]

    def _build_loss_function(self):
        self.loss = Graph2SeqLoss(vocab=self.vocab.in_word_vocab,
                                  use_coverage=self.use_coverage, coverage_weight=0.3)

    def train(self):
        max_score = -1
        self._best_epoch = -1
        for epoch in range(200):
            self.model.train()
            self.train_epoch(epoch, split="train")
            self._adjust_lr(epoch)
            if epoch >= 0:
                score = self.evaluate(split="test")
                if score >= max_score:
                    self.logger.info("Best model saved, epoch {}".format(epoch))
                    self.save_checkpoint("best.pth")
                    self._best_epoch = epoch
                max_score = max(max_score, score)
            if epoch >= 30 and self._stop_condition(epoch):
                break
        return max_score

    def _stop_condition(self, epoch, patience=20):
        return epoch > patience + self._best_epoch

    def _adjust_lr(self, epoch):
        def set_lr(optimizer, decay_factor):
            for group in optimizer.param_groups:
                group['lr'] = group['lr'] * decay_factor

        epoch_diff = epoch - self.opt["lr_start_decay_epoch"]
        if epoch_diff >= 0 and epoch_diff % self.opt["lr_decay_per_epoch"] == 0:
            if self.opt["learning_rate"] > self.opt["min_lr"]:
                set_lr(self.optimizer, self.opt["lr_decay_rate"])
                self.opt["learning_rate"] = self.opt["learning_rate"] * self.opt["lr_decay_rate"]
                self.logger.info("Learning rate adjusted: {:.5f}".format(self.opt["learning_rate"]))

    def train_epoch(self, epoch, split="train"):
        assert split in ["train"]
        self.logger.info("Start training in split {}, Epoch: {}".format(split, epoch))
        loss_collect = []
        dataloader = self.train_dataloader
        step_all_train = len(dataloader)
        for step, data in enumerate(dataloader):
            graph_list, tgt, gt_str = data
            tgt = tgt.to(self.device)
            oov_dict = None
            if self.use_copy:
                oov_dict, tgt = prepare_ext_vocab(graph_list, self.vocab, gt_str=gt_str, device=self.device)

            prob, enc_attn_weights, coverage_vectors = self.model(graph_list, tgt, oov_dict=oov_dict)
            loss = self.loss(logits=prob, label=tgt, enc_attn_weights=enc_attn_weights,
                             coverage_vectors=coverage_vectors)
            loss_collect.append(loss.item())
            if step % self.opt["loss_display_step"] == 0 and step != 0:
                self.logger.info("Epoch {}: [{} / {}] loss: {:.3f}".format(epoch, step, step_all_train,
                                                                           np.mean(loss_collect)))
                loss_collect = []
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

    def evaluate(self, split="val"):
        self.model.eval()
        pred_collect = []
        gt_collect = []
        assert split in ["val", "test"]
        dataloader = self.val_dataloader if split == "val" else self.test_dataloader
        for data in dataloader:
            graph_list, tgt, gt_str = data
            if self.use_copy:
                oov_dict = prepare_ext_vocab(graph_list=graph_list, vocab=self.vocab, device=self.device)
                ref_dict = oov_dict
            else:
                oov_dict = None
                ref_dict = self.vocab.out_word_vocab

            prob, _, _ = self.model(graph_list, oov_dict=oov_dict)
            pred = prob.argmax(dim=-1)

            pred_str = wordid2str(pred.detach().cpu(), ref_dict)
            pred_collect.extend(pred_str)
            gt_collect.extend(gt_str)

        score = self.metrics[0].calculate_scores(ground_truth=gt_collect, predict=pred_collect)
        self.logger.info("Evaluation accuracy in `{}` split: {:.3f}".format(split, score))
        return score

    def translate(self):
        self.model.eval()

        pred_collect = []
        gt_collect = []
        dataloader = self.test_dataloader
        for data in dataloader:
            graph_list, tgt, gt_str = data
            if self.use_copy:
                oov_dict = prepare_ext_vocab(graph_list=graph_list, vocab=self.vocab, device=self.device)
                ref_dict = oov_dict
            else:
                oov_dict = None
                ref_dict = self.vocab.out_word_vocab

            pred = self.model.translate(graph_list=graph_list, oov_dict=oov_dict, beam_size=3, topk=1)

            pred_ids = pred[:, 0, :]  # we just use the top-1

            pred_str = wordid2str(pred_ids.detach().cpu(), ref_dict)

            pred_collect.extend(pred_str)
            gt_collect.extend(gt_str)

        score = self.metrics[0].calculate_scores(ground_truth=gt_collect, predict=pred_collect)
        self.logger.info("Evaluation accuracy in `{}` split: {:.3f}".format("test", score))
        return score

    def load_checkpoint(self, checkpoint_name):
        checkpoint_path = os.path.join(self.opt["checkpoint_save_path"], checkpoint_name)
        self.model.load_state_dict(torch.load(checkpoint_path))

    def save_checkpoint(self, checkpoint_name):
        checkpoint_path = os.path.join(self.opt["checkpoint_save_path"], checkpoint_name)
        if not os.path.exists(self.opt["checkpoint_save_path"]):
            os.makedirs(self.opt["checkpoint_save_path"], exist_ok=True)
        torch.save(self.model.state_dict(), checkpoint_path)


if __name__ == "__main__":
    opt = get_args()
    runner = Jobs(opt)
    # max_score = runner.train()
    # runner.logger.info("Train finish, best val score: {:.3f}".format(max_score))
    runner.load_checkpoint("best.pth")
    runner.translate()



import argparse

def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--root_dir", type=str, default="examples/pytorch/summarization/cnn",
                        help="")
    parser.add_argument("--topology_subdir", type=str, default='DependencyGraph_3w', help="")
    parser.add_argument('--word-emb-size', type=int, default=128, help='')
    parser.add_argument("--log-file", type=str, default="examples/pytorch/summarization/cnn/log/log_dep_30k.txt")
    parser.add_argument("--checkpoint-save-path", type=str, default="examples/pytorch/summarization/cnn/save_dep_30k")
    parser.add_argument('--hidden-size', type=int, default=512, help='')
    parser.add_argument('--dropout', type=float, default=0.2, help='')
    parser.add_argument('--learning-rate', type=float, default=0.001, help='')
    parser.add_argument("--loss-display-step", type=int, default=100, help=' ')
    parser.add_argument("--eval-display-number", type=int, default=3, help="")
    parser.add_argument("--lr-start-decay-epoch", type=int, default=4, help="")
    parser.add_argument("--lr-decay-rate", type=float, default=0.9)
    parser.add_argument("--lr-decay-per-epoch", type=int, default=5, help="")
    parser.add_argument("--min-lr", type=float, default=1e-5, help="")
    parser.add_argument("--use_gpu", type=float, default=1, help="0 for don't use cuda, 1 for using cuda")
    parser.add_argument("--gpu", type=int, default=0, help="gpu id")
    parser.add_argument("--seed", type=int, default=1236, help="")

    parser.add_argument("--use_copy", type=bool, default=False, help="")
    parser.add_argument("--use_coverage", type=bool, default=False, help="")

    parser.add_argument("--batch_size", type=int, default=30, help="batch size")
    parser.add_argument("--gnn", type=str, default="GraphSage", help="")
    parser.add_argument("--direction_option", type=str, default="undirected", help="")
    parser.add_argument("--rnn_dropout", type=float, default=0.2, help="rnn dropout")
    parser.add_argument("--word_dropout", type=float, default=0.2, help="word dropout")
    cfg = parser.parse_args()
    return cfg

# screen -S ghn_code1 python -m examples.pytorch.summarization.cnn.main_g2s --word-emb-size 128 --topology_subdir DependencyGraph_3w --hidden-size 512 --batch_size 40 --log-file examples/pytorch/summarization/cnn/log/log_g2s_30w_copy.txt --checkpoint-save-path /raid/ghn/graph4nlp/examples/pytorch/summarization/cnn/save_g2s_3w_copy --lr-decay-rate 0.8 --learning-rate 1e-3 --use_copy True --use_coverage True
# screen -S ghn_code2 python -m examples.pytorch.summarization.cnn.main_g2s --word-emb-size 128 --topology_subdir DependencyGraph_9w --hidden-size 512 --batch_size 40 --log-file examples/pytorch/summarization/cnn/log/log_g2s_9w_copy.txt --checkpoint-save-path /raid/ghn/graph4nlp/examples/pytorch/summarization/cnn/save_g2s_9w_copy --lr-decay-rate 0.8 --learning-rate 1e-3 --use_copy True --use_coverage True

# --root_dir
# .
# --word-emb-size
# 128
# --topology_subdir
# Seq2seq_30k_t
# --hidden-size
# 512
# --batch_size
# 40
# --log-file
# log/log_s2s_30k_copy3.txt
# --checkpoint-save-path
# save_s2s_30k_copy3
# --lr-decay-rate
# 0.8
# --learning-rate
# 0.001
# --use_copy
# True
# --use_coverage
# True



import os
import torch
import json
from nltk.tokenize import word_tokenize
from graph4nlp.pytorch.data.dataset import KGCompletionDataset, KGDataItem
from ..modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from ..modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
from ..data.data import GraphData
from ..modules.utils.vocab_utils import Vocab
import numpy as np

dataset_root = '../test/dataset/kinship'


class KinshipDataset(KGCompletionDataset):
    @property
    def raw_file_names(self) -> dict:
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'e1rel_to_e2_ranking_train.json',
                'val': 'e1rel_to_e2_ranking_dev.json',
                'test': 'e1rel_to_e2_ranking_test.json'}

    @property
    def processed_file_names(self) -> dict:
        return {'vocab': 'vocab.pt', 'data': 'data.pt', 'KG_graph': 'KG_graph.pt'}

    def download(self):
        # raise NotImplementedError(
        #     'This dataset is now under test and cannot be downloaded. Please prepare the raw data yourself.')
        return

    def __init__(self, root_dir, topology_builder=None, topology_subdir=None, edge_strategy=None, **kwargs):
        self.split_token = ' '
        super(KinshipDataset, self).__init__(root_dir=root_dir, topology_builder=topology_builder,
                                             topology_subdir=topology_subdir,edge_strategy=edge_strategy, **kwargs)


if __name__ == '__main__':
    kinshipdataset = KinshipDataset(root_dir='/Users/gaohanning/PycharmProjects/graph4nlp/examples/pytorch/kg_completion/kinship',
                                    topology_builder=None,
                                    topology_subdir='e1rel_to_e2')

    a = 0
